#include <stdio.h>
#include "enclave_u.h"

#ifdef _MSC_VER
#else
#include <sys/sysinfo.h>
#include <unistd.h>
#endif

#include "sgx_eid.h"
#include "sgx_urts.h"
#include "sgx_utils.h"
#include "sgx_capable.h"


#ifndef _ENCLAVE_MANAGER_
#define _ENCLAVE_MANAGER_


#ifdef _MSC_VER
#define ENCLAVE_NAME "Enclave.signed.dll"
#else
#define ENCLAVE_NAME "enclave.signed.so"
#endif

  
/* Apply the constructor attribute so that it is executed before main() */
void create_enclave (void) __attribute__ ((constructor)); 
  
/* Apply the destructor attribute so that it is executed after main() */
void destroy_enclave (void) __attribute__ ((destructor)); 
  


sgx_enclave_id_t enclave_id = 0;

/* create the enclave, exit there is problems */
/* return 0 if ok, exit -1 for compatibility problem, exit -2 enclave load problem */

int create_enclave(void){

    int update = 0;
    sgx_launch_token_t token = {0};

    if(query_and_enable_sgx_device()==-1){
        //compatibility problem
        //default error handling
        printf("Compatibility problem.\n");
        exit(-1);
    } 

    //destroy the enclave if already exists
    destroy_enclave();
    
    // load enclave
    if (SGX_SUCCESS != sgx_create_enclave(ENCLAVE_NAME, SGX_DEBUG_FLAG, &token, &update, &enclave_id, NULL)) {
        destroy_enclave();
        
        //default error handling
        printf("failed to load enclave.\n");
        exit(-2);
    }

    // succeed to load enclaves. 
    return 0;
}

/* return 0 if ok,-1  if error */ 
int destroy_enclave(void){
    if(enclave_id == 0)
        return 0;

    if(SGX_SUCCESS != sgx_destroy_enclave(enclave_id))
        exit(-1);
	
	enclave_id = 0;
    //manage the error if needed
}



/* name: query_and_enable_sgx_device()
 * Usage: 1. It detects SGX device's status inside target environment.
 *        2. And it will enable SGX device when it detects that SGX was set as Software Control inside BIOS. 
 *           So it can be used inside desktop application installer to enable SGX in target environment during installation.
 * Note: This function need be executed with Administrator privilege.
*/
int query_and_enable_sgx_device()
{
	sgx_device_status_t sgx_device_status;
	sgx_status_t sgx_ret = sgx_cap_get_status(&sgx_device_status);
	if (sgx_ret != SGX_SUCCESS) {
		printf("Failed to get SGX device status. If you have problem, run with administrator privilege\n");
		return -2;
	}
	else {
		switch (sgx_device_status) {
		case SGX_ENABLED:
			printf("SGX device has been enabled.\n");
            return 0; //ok
		
        case SGX_DISABLED_REBOOT_REQUIRED:
			printf("SGX device has been enabled. Please reboot your machine.\n");
			break;
		case SGX_DISABLED_LEGACY_OS:
			printf("SGX device can't be enabled on an OS that doesn't support EFI interface.\n");
			break;
		case SGX_DISABLED:
			printf("SGX is not enabled on this platform. More details are unavailable.\n");
			break;
		case SGX_DISABLED_MANUAL_ENABLE:
			printf("SGX device can be enabled manually in the BIOS setup.\n");
			break;
		case SGX_DISABLED_HYPERV_ENABLED:
			printf("Detected an unsupported version of Windows* 10 with Hyper-V enabled.\n");
			break;
		case SGX_DISABLED_UNSUPPORTED_CPU:
			printf("SGX is not supported by this CPU.\n");
			break;
		default:
			printf("Unexpected error.\n");
			break;

		case SGX_DISABLED_SCI_AVAILABLE:
			printf("SGX device can be enabled by a Software Control Interface, so enable it\n");
			sgx_ret = sgx_cap_enable_device(&sgx_device_status);
			if (sgx_ret != SGX_SUCCESS) {
				printf("Failed to Enable SGX device.\n");
			}
			switch (sgx_device_status)
			{
			case SGX_ENABLED:
				printf("SGX device has been enabled by Software Control Interface.\n");
				return 0; //ok

			case SGX_DISABLED_REBOOT_REQUIRED:
				printf("SGX device has been enabled by Software Control Interface. Please reboot your machine.\n");
				break;
			case SGX_DISABLED_LEGACY_OS:
				printf("SGX device can't be enabled on an OS that doesn't support EFI interface.\n");
				break;
			case SGX_DISABLED:
				printf("SGX is not enabled on this platform. More details are unavailable.\n");
				break;
			case SGX_DISABLED_MANUAL_ENABLE:
				printf("SGX device can be enabled manually in the BIOS setup.\n");
				break;
			case SGX_DISABLED_HYPERV_ENABLED:
				printf("Detected an unsupported version of Windows* 10 with Hyper-V enabled.\n");
				break;
			case SGX_DISABLED_UNSUPPORTED_CPU:
				printf("SGX is not supported by this CPU.\n");
				break;
			default:
				printf("Unexpected error.\n");
                break;
			}
			break;
		}
	}
	return -1;
}

#endif /* !_ENCLAVE_MANAGER_ */




  
